/*
 * Copyright (c), Lawrence Foard, 1992
 *
 * Developed under contract by Virtual World Information systems.
 * Questions and suggestions about this package may be directed to:
 * lfoard@turing.org, entropy@world.std.com
 * 
 * Those without e-mail access can call the VWIS Linux support BBS
 * at 1-508-793-9568 and send e-mail to entropy,
 * or send snail mail to:
 *
 * Virtual World Information Systems
 * 217 Park Ave.
 * Box #108
 * Worcester MA 01609
 *
 * This code may be used under the terms of the GNU license.
 * Other licensing agreements may be negotiated with VWIS.
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <linux/stat.h>
#include "head.h"

char *kernel_stack;
int call_stack_pointer;
extern int errno,sig_count;
extern unsigned char sig_counts[32];

PROCESS *proc;

int do_exec(PROCESS *process,int i_name,int i_argv,int i_envp)
 {
  char *name=&process->prog_area[i_name];
  int a;
  int *argv=(int *)&process->prog_area[i_argv];
  char **nargv;
  int *envp=(int *)&process->prog_area[i_envp];
  char **nenvp;
  for(a=0;argv[a];a++);
  nargv=(char **) malloc(sizeof(char *)*(a+1));
  for(a=0;argv[a];a++)
   {
    char *p,*s;
    int len;
    nargv[a]=&process->prog_area[argv[a]];
    if (process->debug)
     {
      fputs(nargv[a],process->debug);
      fputc(' ',process->debug);
     }
   }
  if (process->debug)
   fputc('|',process->debug);
  nargv[a]=NULL;
  for(a=0;envp[a];a++);
  nenvp=(char **) malloc(sizeof(char *)*(a+1));
  for(a=0;envp[a];a++)
   {
    char *p,*s;
    int len;
    nenvp[a]=&process->prog_area[envp[a]];
    if (process->debug)
     {
      fputs(nenvp[a],process->debug);
      fputc(' ',process->debug);
     }
   }
  if (process->debug)
   {
    fputc('\n',process->debug);
    fflush(process->debug);
   }
  nenvp[a]=NULL;
  a=execve(name,nargv,nenvp);
  free(nargv);
  free(nenvp);
  return a;
 }

/* intercept "system calls" to local selector 0 */

#define CK(a) if (-1==(regs->eax=(a))) {regs->eax=errno;SET_CARRY(regs);}
#define RET ((regs->flags & FLAG_CF) ? -regs->eax : regs->eax)

void system_call()
 {
  int a,old;
  struct registers *regs;
  struct stack_top *stop;
  char *p;
  char buff[1000];
  regs=&(stop=(struct stack_top *)(call_stack_pointer+proc->prog_area))->regs;
  CLEAR_CARRY(regs);
  switch(regs->eax)
   { 
    case 0x01:
     if (proc->debug)
      {
       fprintf(proc->debug,"exit(%d)\n",stop->stack[0]);
       fflush(proc->debug);
      }
     exit(stop->stack[0]);
     break;
    case 0x02: /* fork */
     CK(fork());
     regs->edx=(regs->eax) ? 0 : getppid();
     if (proc->debug) 
      {
       fprintf(proc->debug,"%d=fork()\n",RET);
       fflush(proc->debug);
      }
     break;
    case 3: /* read */
     regs->eax=read(stop->stack[0],stop->stack[1]+proc->prog_area,stop->stack[2]);
     if (regs->eax==-1)
      {
       SET_CARRY(regs);
       regs->eax=errno;
      }
     if (proc->debug) 
      {
       fprintf(proc->debug,"%d=read(%d,%d,%d)\n",RET,stop->stack[0],stop->stack[1],stop->stack[2]);
       fflush(proc->debug);
      }
     break;
    case 4: /* write */
     CK(write(stop->stack[0],stop->stack[1]+proc->prog_area,stop->stack[2]));
     if (proc->debug)
      {
       fprintf(proc->debug,"%d=write(%d,%d,%d)\n",RET,stop->stack[0],stop->stack[1],stop->stack[2]);
       fflush(proc->debug);
      }
     break;
    case 0x5: /* open */
     CK(open(p=&proc->prog_area[stop->stack[0]],xtol(stop->stack[1]),stop->stack[2]));
     if (proc->debug) 
      {
       int a;
       fprintf(proc->debug,"%d=open(\"",RET);
       for(a=0;p[a];a++)
        fputc(p[a],proc->debug);
       fprintf(proc->debug,"\",%d,%d)\n",stop->stack[1],stop->stack[2]);
       fflush(proc->debug);
      }
     break;
    case 0x6: /* close */
     CK(close(stop->stack[0]));
     if (proc->debug)
      {
       fprintf(proc->debug,"%d=close(%d)\n",RET,stop->stack[0]);
       fflush(proc->debug);
      }
     break;   
    case 0x7: /* wait */
     {
      int a;
      if (proc->debug)
       {
        fprintf(proc->debug,"entering wait\n");
        fflush(proc->debug);
       }
       CK(wait(&proc->prog_area[stop->stack[0]]));
      if (proc->debug)
       {
        fprintf(proc->debug,"%d=wait([data area])\n",RET);
        fflush(proc->debug);
       }
     }
     break;
    case 0x08: /* create */
     CK(creat(p=&proc->prog_area[stop->stack[0]],stop->stack[1]));
     if (proc->debug) 
      {
       int a;
       fprintf(proc->debug,"%d=creat(\"",RET);
       for(a=0;p[a];a++)
        fputc(p[a],proc->debug);
       fprintf(proc->debug,"\",%d)\n",stop->stack[1]);
       fflush(proc->debug);
      }
     break;
    case 0x0c: /* chdir */
     CK(chdir(p=&proc->prog_area[stop->stack[0]]));
     if (proc->debug) 
      {
       int a;
       fprintf(proc->debug,"%d=chdir(\"",RET);
       for(a=0;p[a];a++)
        fputc(p[a],proc->debug);
       fprintf(proc->debug,"\")\n");
       fflush(proc->debug);
      }
     break;
    case 0x0d: /* time */
     regs->eax=(stop->stack[0]) ? time(&proc->prog_area[stop->stack[0]]) : time(NULL);
     if (proc->debug)
      {
       fprintf(proc->debug,"%d=time([data area])\n",RET);
       fflush(proc->debug);
      }
     break;
    case 0x11: /* break */
     if (proc->debug)  
      {
       fprintf(proc->debug,"%d=brk(%d)\n",0,stop->stack[0]);
       fflush(proc->debug);
      }
     regs->eax=0;
     proc->seg_break=stop->stack[0];
     break;
    case 0x12: /* stat */
     { 
      struct new_stat st;
      CK(stat(&proc->prog_area[stop->stack[0]],&st));
      lst_to_xst(&st,&proc->prog_area[stop->stack[1]]);
      if (proc->debug) 
       {
        int a;
        char *p;
        fprintf(proc->debug,"%d=stat(\"",RET);
        p=&proc->prog_area[stop->stack[0]];
        for(a=0;p[a];a++)
         fputc(p[a],proc->debug);
        fprintf(proc->debug,"\",%d)\n",stop->stack[1]);
        fflush(proc->debug);
       }
      }
     break;
    case 0x13: /* lseek */
     CK(lseek(stop->stack[0],stop->stack[1],stop->stack[2]));
     if (proc->debug) 
      {
       fprintf(proc->debug,"%d=lseek(%d,%d,%d)\n",RET,stop->stack[0],stop->stack[1],stop->stack[2]);
       fflush(proc->debug);
      }
     break;
    case 0x14: /* getpid */
     CK(getpid());
     if (proc->debug)
      {
       fprintf(proc->debug,"%d=getpid()\n",RET);
       fflush(proc->debug);
      }
     break;
    case 0x18: /* geteuid */
     CK(geteuid());
     if (proc->debug) 
      {
       fprintf(proc->debug,"%d=geteuid()\n",RET);
       fflush(proc->debug);
      }
     break; 
    case 0x1b: /* alarm */
     CK(alarm(stop->stack[0]));
     if (proc->debug) 
      {
       fprintf(proc->debug,"%d=alarm(%d)\n",RET,stop->stack[0]);
       fflush(proc->debug);
      }
     break; 
    case 0x1c: /* fstat */
     {
      struct new_stat st;
      CK(fstat(stop->stack[0],&st));
      lst_to_xst(&st,&proc->prog_area[stop->stack[1]]);
      if (proc->debug) 
       {
        fprintf(proc->debug,"%d=fstat(%d,[data])\n",RET,stop->stack[0]);
        fflush(proc->debug);
       }
      break;
     }
    case 0x2a: /* pipe */
     {
      int fds[2];
      if (-1==pipe(fds))
       {
        SET_CARRY(regs);
        regs->eax=errno;
       }
      else
       {
        regs->eax=fds[0];
        regs->edx=fds[1];
       }
      if (proc->debug) 
       {
        fprintf(proc->debug,"%d=pipe({%d,%d})\n",RET,fds[0],fds[1]);
        fflush(proc->debug);
       }
     }
     break;
    case 0x2b: /* times */
     {
      int *p;
      int buff[4];
      p=&proc->prog_area[stop->stack[0]];
     CK(times(buff));
     p[0]=buff[0];
     p[1]=buff[1];
     p[2]=buff[2];
     p[3]=buff[3];
     if (proc->debug) 
      {
       fprintf(proc->debug,"%d=times()\n",RET);
       fflush(proc->debug);
      }
     }
     break;
    case 0x2f: /* getegid */
     CK(getegid());
     if (proc->debug) 
      {
       fprintf(proc->debug,"%d=getegid()\n",RET);
       fflush(proc->debug);
      }
     break; 
    case 0x30: /* sigcomm */
     switch(0x1f00 & stop->stack[0])
      {
       case 0:
        CK(set_sig(stop->stack[0],stop->stack[1]));
       default:
        regs->eax=0;
      }
     /* 0,signum | 0x200 hold,0x1000 pause,0x400 release,0x800 ignore */
     if (proc->debug) 
      {
       fprintf(proc->debug,"_sigcomm(%d,%d,%d)\n",stop->stack[0],stop->stack[1],stop->stack[2]);
       fflush(proc->debug);
      }
     break;
    case 0x36: /* ioctl */
     CK(do_ioctl(proc,stop->stack[0],stop->stack[1],stop->stack[2]));
     if (proc->debug) 
      {
       fprintf(proc->debug,"%d=ioctl(%d,0x%4.4x,%d)\n",RET,stop->stack[0],stop->stack[1],stop->stack[2]);
       fflush(proc->debug);
      }
     break;
    case 0x3b: /* execve */
     if (proc->debug)
      {
       int a;
       char *p;
       fprintf(proc->debug,"?=execve(\"");
       p=&proc->prog_area[stop->stack[0]];
       fputs(p,proc->debug);
       fprintf(proc->debug,"\",args,envp)\n");
       fflush(proc->debug);
      }
     CK(do_exec(proc,stop->stack[0],stop->stack[1],stop->stack[2]));
     if (proc->debug)
      {
       fprintf(proc->debug,"execve failed %d\n",RET);
       fflush(proc->debug);
      }
     break;
    case 0x3e: /* fcntl */
     CK(do_fcntl(proc,stop->stack[0],stop->stack[1],stop->stack[2]));
     if (proc->debug) 
      {
       fprintf(proc->debug,"%d=fcntl(%d,0x%4.4x,%d)\n",RET,stop->stack[0],stop->stack[1],stop->stack[2]);
       fflush(proc->debug);
      }
     break;
    case 0x128: /* locking */
     regs->eax=0;
     
     break;
    case 0x728: /* rdchk */
     {
      int rd=1<<stop->stack[0];
      struct timeval ti;
      ti.tv_sec=0;
      ti.tv_usec=0;
      CK(select(stop->stack[0]+1,&rd,NULL,NULL,&ti));
      if (proc->debug) 
       {
        fprintf(proc->debug,"%d=rdchk(%d)\n",RET,stop->stack[0]);
        fflush(proc->debug);
       }
     }
     break;
    case 0xb28: /* ftime */
     regs->eax=1;
     SET_CARRY(regs);
     if (proc->debug)
      {
       fprintf(proc->debug,"%d=ftime()\n",RET);
       fflush(proc->debug);
      }
     break;
    default: 
     fprintf(stderr,"call %x 3=%d 2=%d 1=%d 0=%d\n",regs->eax,stop->stack[3],stop->stack[2],stop->stack[1],stop->stack[0]);
     exit(1);     
     break;
   }
 }

/*extern void called();*/

void called1()
 {
  write(1,"dfklgjklfg\n\n\n\n\n\n\n\n\n\n",15);
 /* _exit(0);*/
 }

void (*initial)();

int enn;

void help()
 {
  fprintf(stderr,"Emul Version %s Copyright (c), Lawrence Foard, 1992\n\n",VERSION);
  fprintf(stderr,"May be used and distributed under the GNU license.\n");  
  fprintf(stderr,"Xenix (tm)* system call emulator written under contract\n");
  fprintf(stderr,"by Virtual World Information Systems.\n");
  fprintf(stderr,"To report bugs or obtain the latest version write to:\n");
  fprintf(stderr,"entropy@world.std.com,lfoard@turing.org, call the VWIS Linux support BBS\n");
  fprintf(stderr,"1-508-793-9568 and send e-mail to entropy, or send snail mail\n");
  fprintf(stderr,"to:\n\n");
  fprintf(stderr,"Virtual World Information Systems\n");
  fprintf(stderr,"217 Park Ave.\n");
  fprintf(stderr,"Box #108\n");
  fprintf(stderr,"Worcester MA 01609\n\n");
  fprintf(stderr,"Please consider VWIS for your Linux and C contracting needs.\n");
  fprintf(stderr,"We offer contract software help as well as support for Linux.\n\n");
  fprintf(stderr,"* Xenix is a trade mark of The Santa Crux Operation\n");
 }
 
main(int argc,char *argv[],char *envp[])
 {
  char a,farg;
  char *tmp;
  int u;
  int stack_size=STACK_SIZE;
  int heap_size=HEAP_SIZE;
  char *debug=NULL;
  PROCESS *process;
  for(a=1;argv[a] && (*argv[a]=='-');a++)
   switch(argv[a][1])
    {
     case 'h':
      help();
      exit(0);
     case 'b':
      if (!argv[++a])
       exit(1);
      heap_size=1024*atoi(argv[a]);
      break;
     case 's':
      if (!argv[++a])
       exit(1);
      stack_size=1024*atoi(argv[a]);
      break;
     case 'd':
      if (!argv[++a])
       exit(1);
      debug=argv[a];
      break;
     default:
      fprintf(stderr,"Xenix (tm) system call emulator Copyright (c), Lawrence Foard, 1992\n");
      fprintf(stderr,"Usage: %s [-b heap_size (K)] [-s stack_size (K)] [-d debug_log_file]\n   [-l (license info)] [Xenix program]\n",argv[0]);
      fprintf(stderr,"Default stack size=%dK heap size=%dK\n",STACK_SIZE/1024,HEAP_SIZE/1024);
      exit(1);
    }
  farg=a;
  for(a=0;a<32;a++)
   sig_counts[a]=0;
  init_cvt();
  process=load_file(argv[farg],(debug) ? fopen(debug,"w") : NULL,stack_size,heap_size);
  /* initialize the "kernel" stack */
  kernel_stack=(&a)-10000;
  /* create segments for the xenix program */
  set_segments(process);
  /* set global process variable (yuck) */
  proc=process;
  initial=0;
  /* set arguments */
  set_argvnc(process,argv+farg,envp);
  /* initial xenix stack pointer */
  call_stack_pointer=process->user_stack;
  /* initial heap */
  enn=process->seg_break;
  /* jump into xenix program */
  startit();
  _exit(0);
 }
